@@ -1,8 +1,10 @@
 import json
+import os
 import urllib.parse
 from collections import OrderedDict
 from contextlib import contextmanager
-from typing import Any, Dict, Generator, Optional
+from pathlib import Path
+from typing import Any, Dict, Generator, List, Optional, Union
 
 import docker
 import requests
@@ -12,10 +14,14 @@ from tqdm import tqdm
 
 from .config import Config
 from .exceptions import *
+from .version import VERSION
 
 __all__ = ["Beaker"]
 
 
+PathOrStr = Union[os.PathLike, Path]
+
+
 class Beaker:
     """
     A client for interacting with `Beaker <https://beaker.org>`_.
@@ -28,7 +34,7 @@ class Beaker:
     def __init__(self, config: Config):
         self.config = config
         self.base_url = f"{self.config.agent_address}/api/{self.API_VERSION}"
-        self.docker = docker.from_env()
+        self._docker: Optional[docker.DockerClient] = None
 
     @property
     def user(self) -> str:
@@ -38,12 +44,18 @@ class Beaker:
         return self.whoami()["name"]
 
     @classmethod
-    def from_env(cls, **kwargs) -> "Beaker":
+    def from_env(cls, **overrides) -> "Beaker":
         """
-        Initialize client from environment variables. Expects the beaker auth token
-        to be set as the ``BEAKER_TOKEN`` environment variable.
+        Initialize client from a config file and/or environment variables.
         """
-        return cls(Config.from_env())
+        return cls(Config.from_env(**overrides))
+
+    @property
+    def docker(self) -> docker.DockerClient:
+        if self._docker is None:
+            self._docker = docker.from_env()
+        assert self._docker is not None
+        return self._docker
 
     @contextmanager
     def _session_with_backoff(self) -> requests.Session:
@@ -61,20 +73,26 @@ class Beaker:
         resource: str,
         method: str = "GET",
         query: Optional[Dict[str, Any]] = None,
-        data: Optional[Dict[str, Any]] = None,
+        data: Optional[Any] = None,
         exceptions_for_status: Optional[Dict[int, BeakerError]] = None,
+        headers: Optional[Dict[str, str]] = None,
+        token: Optional[str] = None,
+        base_url: Optional[str] = None,
     ) -> requests.Response:
         with self._session_with_backoff() as session:
-            url = f"{self.base_url}/{resource}"
+            url = f"{base_url or self.base_url}/{resource}"
             if query is not None:
                 url = url + "?" + urllib.parse.urlencode(query)
+            default_headers = {
+                "Authorization": f"Bearer {token or self.config.user_token}",
+                "Content-Type": "application/json",
+            }
+            if headers is not None:
+                default_headers.update(headers)
             response = getattr(session, method.lower())(
                 url,
-                headers={
-                    "Authorization": f"Bearer {self.config.user_token}",
-                    "Content-Type": "application/json",
-                },
-                data=None if data is None else json.dumps(data),
+                headers=default_headers,
+                data=json.dumps(data) if isinstance(data, dict) else data,
             )
             if exceptions_for_status is not None and response.status_code in exceptions_for_status:
                 raise exceptions_for_status[response.status_code]
@@ -105,6 +123,23 @@ class Beaker:
             exceptions_for_status={404: WorkspaceNotFound(workspace_name)},
         ).json()
 
+    def ensure_workspace(self, workspace: str):
+        """
+        Ensure the given workspace exists.
+
+        Raises
+        ------
+        HTTPError
+        ValueError
+            If the workspace name is invalid.
+
+        """
+        try:
+            self.get_workspace(workspace)
+        except WorkspaceNotFound:
+            org, name = workspace.split("/")
+            self.request("workspaces", method="POST", data={"name": name, "org": org})
+
     def get_experiment(self, exp_id: str) -> Dict[str, Any]:
         """
         Get info about an experiment.
@@ -116,7 +151,8 @@ class Beaker:
 
         """
         return self.request(
-            f"experiments/{exp_id}", exceptions_for_status={404: ExperimentNotFound(exp_id)}
+            f"experiments/{urllib.parse.quote(exp_id, safe='')}",
+            exceptions_for_status={404: ExperimentNotFound(exp_id)},
         ).json()
 
     def create_experiment(
@@ -134,6 +170,8 @@ class Beaker:
         workspace_name = workspace or self.config.default_workspace
         if workspace_name is None:
             raise ValueError("'workspace' argument required")
+        # Ensure workspace exists.
+        self.ensure_workspace(workspace_name)
         return self.request(
             f"workspaces/{urllib.parse.quote(workspace_name, safe='')}/experiments",
             method="POST",
@@ -153,7 +191,80 @@ class Beaker:
 
         """
         return self.request(
-            f"datasets/{dataset_id}", exceptions_for_status={404: DatasetNotFound(dataset_id)}
+            f"datasets/{urllib.parse.quote(dataset_id, safe='')}",
+            exceptions_for_status={404: DatasetNotFound(dataset_id)},
+        ).json()
+
+    def delete_dataset(self, dataset_id: str):
+        self.request(
+            f"datasets/{urllib.parse.quote(dataset_id, safe='')}",
+            method="DELETE",
+            exceptions_for_status={404: DatasetNotFound(dataset_id)},
+        )
+
+    def create_dataset(
+        self,
+        name: str,
+        source: PathOrStr,
+        target: Optional[str] = None,
+        workspace: Optional[str] = None,
+        force: bool = False,
+    ) -> Dict[str, Any]:
+        """
+        Create a dataset with the source file(s).
+        """
+        workspace_name = workspace or self.config.default_workspace
+        if workspace_name is None:
+            raise ValueError("'workspace' argument required")
+
+        # Ensure workspace exists.
+        self.ensure_workspace(workspace_name)
+
+        # Ensure source exists.
+        source: Path = Path(source)
+        if not source.exists():
+            raise FileNotFoundError(source)
+
+        if not source.is_file():
+            raise NotImplementedError("'create_dataset()' only works for single files so far")
+
+        # Create the dataset.
+        def make_dataset() -> Dict[str, Any]:
+            return self.request(
+                "datasets",
+                method="POST",
+                query={"name": name},
+                data={"workspace": workspace_name, "fileheap": True},
+                exceptions_for_status={409: DatasetConflict(name)},
+            ).json()
+
+        try:
+            dataset_info = make_dataset()
+        except DatasetConflict:
+            if force:
+                self.delete_dataset(f"{self.user}/{name}")
+                dataset_info = make_dataset()
+            else:
+                raise
+
+        # Upload the file.
+        with source.open("rb") as source_file:
+            self.request(
+                f"datasets/{dataset_info['storage']['id']}/files/{target or source.name}",
+                method="PUT",
+                data=source_file,
+                token=dataset_info["storage"]["token"],
+                base_url=dataset_info["storage"]["address"],
+                headers={
+                    "User-Agent": f"beaker-py v{VERSION}",
+                },
+            )
+
+        # Commit the dataset.
+        return self.request(
+            f"datasets/{dataset_info['id']}",
+            method="PATCH",
+            data={"commit": True},
         ).json()
 
     def get_logs(self, job_id: str) -> Generator[bytes, None, None]:
@@ -172,7 +283,7 @@ class Beaker:
         content_length = response.headers.get("Content-Length")
         total = int(content_length) if content_length is not None else None
         progress = tqdm(
-            unit="iB", unit_scale=True, unit_divisor=1024, total=total, desc="downloading"
+            unit="iB", unit_scale=True, unit_divisor=1024, total=total, desc="downloading logs"
         )
         for chunk in response.iter_content(chunk_size=1024):
             if chunk:
@@ -189,6 +300,8 @@ class Beaker:
         ------
         ExperimentNotFound
         JobNotFound
+        HTTPError
+
         """
         exp = self.get_experiment(exp_id)
         if job_id is None:
@@ -227,6 +340,11 @@ class Beaker:
 
         """
         workspace = workspace or self.config.default_workspace
+        if workspace is None:
+            raise ValueError("'workspace' argument required")
+
+        # Ensure workspace exists.
+        self.ensure_workspace(workspace)
 
         # Get local Docker image object.
         image = self.docker.images.get(image_tag)
@@ -289,3 +407,12 @@ class Beaker:
             method="DELETE",
             exceptions_for_status={404: ImageNotFound(image_id)},
         )
+
+    def list_experiments(self, workspace: Optional[str] = None) -> List[Dict[str, Any]]:
+        workspace_name = workspace or self.config.default_workspace
+        if workspace_name is None:
+            raise ValueError("'workspace' argument required")
+        return self.request(
+            f"workspaces/{urllib.parse.quote(workspace_name, safe='')}/experiments",
+            exceptions_for_status={404: WorkspaceNotFound(workspace_name)},
+        ).json()["data"]
